#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdio.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <semaphore.h>

#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <queue>
#include <signal.h>

using namespace std;

int server_sock, client_sock;
struct sockaddr_in local_addr, remote_addr;
socklen_t remote_addr_len;

int countBranch;
int globalStatue = 0;
int branchesSocket[2];
std::vector<pthread_t> bank_threads;

struct Branch {
	struct sockaddr_in remote_addr;
	int branchSocket;
	timeval casOdeslani;
	int port;
	int idBranch;
};

struct BranchInfo {
	int port;
	int idBranch;
};

struct RequestGlobalStatue {
	int request;
	timeval casOdeslani;
	timeval casPrijeti;
	int currentMoney;
};

vector<Branch> branches;

void settingBankAsServer(int port) {

	server_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (server_sock < 0) {
		cout << "The problem occurred during creating socket.\n";
		exit(EXIT_FAILURE);
	}

	int reuse = 1;
	setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &reuse,
			sizeof(reuse));

	memset(&local_addr, 0, sizeof(struct sockaddr_in));

	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons(port);
	local_addr.sin_addr.s_addr = INADDR_ANY;

	if (bind(server_sock, (struct sockaddr *) &local_addr,
			sizeof(struct sockaddr_in)) != 0) {
		cout << "The problem occurred during binding.\n";
		exit(EXIT_FAILURE);
	}

	if (listen(server_sock, 5) != 0) {
		cout << "The problem occurred during listen.\n";
		exit(EXIT_FAILURE);
	}

	cout << "Bank has been started...\n";
}

void sendToOtherBranches(int clientSocket) {
	if (!branches.empty()) {
		for (std::vector<Branch>::iterator it = branches.begin();
				it != branches.end(); it++) {
			//cout << "posilam info o pobocce id - " << (*it).idBranch << "\n";
			send(clientSocket, &(*it), sizeof(vector<Branch> ), 0);
		}
	}
}

void sendInfoToBranches() {

	Branch endBranchSignal;
	endBranchSignal.idBranch = -1;

//	unsigned int arrayLenght =
//			(sizeof(branchesSocket) / sizeof(*branchesSocket));

	for (int i = 0; i < countBranch; i++) {
		for (std::vector<Branch>::iterator it = branches.begin();
				it != branches.end(); it++) {
			send(branchesSocket[i], &(*it), sizeof(Branch), 0);
		}
		send(branchesSocket[i], &endBranchSignal, sizeof(Branch), 0);
	}

}

void setNewBranch(Branch newBranch) {

	BranchInfo branchInfo;
	recv(newBranch.branchSocket, &branchInfo, sizeof(BranchInfo), 0);

	countBranch++;
	branchInfo.idBranch = countBranch;

	send(newBranch.branchSocket, &branchInfo, sizeof(BranchInfo), 0);

//	newBranch.remote_addr = htons(branchInfo.port);
	newBranch.idBranch = branchInfo.idBranch;
	newBranch.port = branchInfo.port;

	branches.push_back(newBranch);
}

timeval getTime() {
	timeval cas;
	gettimeofday(&cas, NULL);
	return cas;
}

ofstream bankLog;
string fileLogBank;

string timevalToString(timeval time) {
	stringstream ss;
	ss << time.tv_sec << "." << time.tv_usec;
	return ss.str();
}

string numberToString(int number) {
	stringstream ss;
	ss << number;
	return ss.str();
}

string timeToString(timeval cas) {
	stringstream ss;
	ss << cas.tv_sec << "." << cas.tv_usec;
	return ss.str();
}

void writeToLog(RequestGlobalStatue request, int idBranch) {
	bankLog.open(fileLogBank.c_str(), std::ios_base::app);
	bankLog << "- src: {p: P" << idBranch << ", ts: "
			<< timeToString(request.casOdeslani) << "}\n";
	//<< (int)operation.sendTime.tv_sec - (int)operation.sendTime.tv_usec << "}\n";
	bankLog << "  dst: {p: BANK, ts: " << timeToString(request.casPrijeti) //(int)operation.recvTime.tv_sec - (int)operation.recvTime.tv_usec
			<< "}\n";
	bankLog << "  color: '#FF0000'\n";
	bankLog << "  data: '" << request.request << ";" << request.currentMoney
			<< "'\n";
	bankLog.close();
}

void writeToStartLog(RequestGlobalStatue request, int idBranch) {
	bankLog.open(fileLogBank.c_str(), std::ios_base::app);
	bankLog << "- src: {p: BANK, ts: " << timeToString(request.casOdeslani)
			<< "}\n";
	//<< (int)operation.sendTime.tv_sec - (int)operation.sendTime.tv_usec << "}\n";
	bankLog << "  dst: {p: P" << idBranch << ", ts: "
			<< timeToString(request.casOdeslani) //(int)operation.recvTime.tv_sec - (int)operation.recvTime.tv_usec
			<< "}\n";
	bankLog << "  data: '" << request.request << ";" << 0 << "'\n";
	bankLog.close();
}

void writeGlobalStatu() {
	bankLog.open(fileLogBank.c_str(), std::ios_base::app);
	bankLog << "- src: {p: BANK, ts: " << timeToString(getTime()) << "}\n";
	//<< (int)operation.sendTime.tv_sec - (int)operation.sendTime.tv_usec << "}\n";
	bankLog << "  dst: {p: BANK, ts: " << timeToString(getTime()) //(int)operation.recvTime.tv_sec - (int)operation.recvTime.tv_usec
			<< "}\n";
	bankLog << "  color: '#FF0000'\n";
	bankLog << "  data: 'GLOBAL - 20000'\n";
	bankLog.close();
}

void *processGlobal(void *args) {
	int *pom = (int *) args;
	int sock = *pom;

	RequestGlobalStatue request;
	recv(sock, &request, sizeof(RequestGlobalStatue), 0);
	request.casPrijeti = getTime();
	writeToLog(request, request.request);
	cout << "Received branch statue - " << request.currentMoney << "\n";
	cout << "from id - " << request.request << "\n";
	globalStatue += request.currentMoney;

	return 0;
}

int main(int argc, char* argv[]) {

	if (argc != 2) {
		cout << "./Banka <port bank>\n";
		exit(1);
	}
	countBranch = 0;

//	struct sigaction sigIntHandler;
//	sigIntHandler.sa_handler = finishBank;
//	sigemptyset(&sigIntHandler.sa_mask);
//	sigIntHandler.sa_flags = 0;
//	sigaction(SIGINT, &sigIntHandler, NULL);

//	fileLogSeq = "../log/log_seq.txt";
	settingBankAsServer(atoi(argv[1]));
	fileLogBank = "../log/log_bank.txt";

	while (branches.size() < 3) {
		cout << "waiting for connecting...\n";
		client_sock = accept(server_sock, (struct sockaddr *) &remote_addr,
				&remote_addr_len);

		if (client_sock > 0) {
			cout << "new branch connected...\n";

			Branch newBranch;
			newBranch.remote_addr = remote_addr;
			newBranch.branchSocket = client_sock;

			branchesSocket[branches.size()] = client_sock;

			setNewBranch(newBranch);
		}
	}

	sendInfoToBranches();

	RequestGlobalStatue request;
	string req;
	pthread_t worker;

	while (true) {

		cout << "Please enter an request for global statue (write global):\n";
		getline(cin, req);

		if (req == "global") {
			globalStatue = 0;
			request.request = 1;
			cout << "Waiting for response\n";
			for (std::vector<Branch>::iterator it = branches.begin();
					it != branches.end(); it++) {

				if ((*it).idBranch == 1) {
					request.casOdeslani = getTime();
					//writeToStartLog(request, (*it).idBranch);
					send((*it).branchSocket, &request,
							sizeof(RequestGlobalStatue), 0);
				}
				int *sock = new int;
				*sock = (*it).branchSocket;
				pthread_create(&worker, NULL, processGlobal, sock);
				bank_threads.push_back(worker);
				//sleep(1);

			}
			for (unsigned int i = 0; i < bank_threads.size(); i++) {
				pthread_join(bank_threads[i], NULL);
			}
//			writeGlobalStatu();

			cout << "Global statue - " << globalStatue << "\n";
		} else {
			cout << "Unknown request\n";
		}

	}

	return 0;
}
